#from ctypes import cdll, CDLL, c_char_p, c_byte, POINTER
from ctypes import *
import os
import shutil

class FlashProARM_API():
    def __init__(self, dll_dir, serials): 
        self.__sel_dll_path = dll_dir + "FlashProARM-FPAsel.dll"
        self.__generic_fpa_dll_path = dll_dir + "FlashProARM-FPA1.dll"
        self.__sel_dll = cdll.LoadLibrary(self.__sel_dll_path)
        
        self.__instances = self.__sel_dll.F_OpenInstancesAndFPAs(serials)
        for instance in range(2, self.__instances+1):
            fpa_dll_path = dll_dir + "FlashProARM-FPA{}.dll".format(instance)
            if not os.path.isfile(fpa_dll_path):
                shutil.copy(self.__generic_fpa_dll_path, fpa_dll_path)

    def get_instances(self):
        return self.__instances        

    # MULTI API-DLL FUNCTIONS
    def F_Trace_ON(self):
        """ Activate tracing for subsequent calls to the Multi API-DLL. Log is saved in DLLtrace.txt located
            in the Multi API-DLL directory. When activated, records all API-DLL function calls from the
            application software invoked via the Multi API-DLL. The DLLtrace.txt is overwritten for each new session."""
        self.__sel_dll.F_Trace_ON()

    def F_Trace_OFF(self):
        """ Disable tracing. """
        self.__sel_dll.F_Trace_OFF()

    def F_OpenInstancesAndFPAs(self, FileName):
        """Multi API-DLL scans USB ports for connected FPAs listed in the setup fiie, or input string. 
            When an FPA listed in the setup file is found, the corresponding API-DLL is copied on disk if necessary, and loaded. 
            F_Initialization should be called for each FPA after this function succeeds. """
        return self.__sel_dll.F_OpenInstancesAndFPAs(FileName)

    def F_CloseInstances(self):
        """All FPAs terminate communication with target devices and close target devices according to configuration 
            settings (power setting, adapter line states, etc.). Finally, the USB connections to the
            FPAs are terminated and API-DLL instances are freed. The Multi API-DLL can now be used to
            open a new set of FPAs using the function F_OpenInstancesAndFPAs."""
        return self.__sel_dll.F_CloseInstances()

    def F_Set_FPA_index(self, fpa):
        """ Select desired FPA index to perform specic tasks (access specic API-DLL instance). """
        return self.__sel_dll.F_Set_FPA_index(fpa)

    def F_Get_FPA_index(self):
        """Get current FPA index."""
        return self.__sel_dll.F_Get_FPA_index()
    
    def F_Check_FPA_index(self):
        """ Get current FPA index and check if index is valid. A valid index corresponds to an individual FPA
            that has been opened with F OpenInstancesAndFPAs. If FPA index was set to 0 (all), then this
            function will only return 0, even if no FPAs are open. Does not indicate whether FPA index is
            enabled or disabled. """
        return self.__sel_dll.F_Check_FPA_index()

    def F_LastStatus(self, fpa):
        """Return value from the last function call issued to the specied FPA (API-DLL instance). This
            function is useful when multiple FPAs are being programmed using FPA index 0 (all), but the
            return value was not the same for all FPAs (i.e. 4 FPAs succeeded, 1 FPA failed). When the
            return value is not the same for all FPAs then, then the actual return value from the Multi API-
            DLL will be FPA UNMATCHED RESULTS. To nd out per FPA return values, select each FPA
            index individually using F Set FPA index, and call F LastStatus for each index. """
        return self.__sel_dll.F_LastStatus(fpa)
    
    def F_Multi_DLLTypeVer(self):
        """ Get Multi API-DLL software version number."""
        return self.__sel_dll.F_Multi_DLLTypeVer()

    def F_Get_FPA_SN(self, fpa):
        """Get serial number (SN) of FPA assigned to specied index."""
        return self.__sel_dll.F_Get_FPA_SN(c_byte(fpa))
    
    def F_Get_FPA_Label(self, fpa, label):
        """ Get detailed information of FPA assigned to specied index.
            Input
            YTE fpa : desired FPA index, 1 to 64.
            BYTE *label : pointer to byte array, of at least FPA LABEL SIZE (80 bytes) """
        return self.__sel_dll.F_Get_FPA_Label(fpa, label)     # TODO: check c_types

    def F_GetProgressBar(self, fpa):
        """ Get internal progress counter value for operations done inside Encapsulated Functions.
            When used in conjunction with F GetLastOpCode it allows the user application to keep track
            of progress in the same way that the GangPro-ARM GUI does."""
        return self.__sel_dll.F_GetProgressBar(fpa)

    def F_GetLastOpCode(self, fpa):
        """ Read internal opcode value for sub-operations done inside Encapsulated Functions.
            When used in conjunction with F GetProgressBar it allows the user application to keep track of
            progress in the same way that the GangPro-ARM GUI does."""
        return self.__sel_dll.F_GetLastOpCode(fpa)

    def F_Disable_FPA_index(self, fpa):
        """ Disable desired FPA index to ignore commands. An FPA index is enabled by default. A disabled
            FPA index will ignore commands when the FPA index is set to it, or 0 (all FPAs).
            Input: BYTE fpa : desired FPA index, 1 to 64 """
        self.__sel_dll.F_Disable_FPA_index(fpa)

    def F_Enable_FPA_index(self, fpa):
        """ Enable desired FPA index to accept commands. An FPA index is enabled by default. A disabled
            FPA index will ignore commands when the FPA index is set to it, or 0 (all FPAs)."""
        self.__sel_dll.F_Enable_FPA_index(fpa)

    # GENERIC FUNCTIONS
    def F_Initialization(self):
        """Select FPA to perform operation on using F_Set_FPA_index, index 1 to 64.Use index 0 to
            perform operation on all FPAs (if results differ, use F_LastStatus to get individual results). """
        return self.__sel_dll.F_Initialization()

    def F_Get_Device_Info(self, index):
        """ Get information about selected microcontroller. """
        return self.__sel_dll.F_Get_Device_Info(index)
    
    def F_Reset_Target(self):
        """ Generate short reset pulse on target device's reset line. The length of the reset pulse can be selected
            using conguration options. The ResetTimeIndex is used to select pre-set times, or for custom setup
            select the last index and specify ResetPulseTime, and ResetIdleTime. """
        return self.__sel_dll.F_Reset_Target()

    def F_Report_Message(self):
        """ Get the last report message from the programmer. When any of the DLL functions are called, an
            output message is usually created. The GUI uses these messages to print to the dialog box. The
            user application can use these messages to populate its own dialog box. The last report message
            can be read by user application using these functions."""
        return self.__sel_dll.F_Report_Message()
    
    def F_Set_MCU_Name(self, mcu_name):
        """ Set target microcontroller name. A list of acceptable names can be obtained using the functions
            F_Set_MCU_Family_Group and F_Get_MCU_Name_list."""
        return self.__sel_dll.F_Set_MCU_Name(mcu_name)

    def F_GetReportMessageChar(self, index):
        """ Get one character from the report message buffer."""
        return self.__sel_dll.F_GetReportMessageChar(index)
    
    def F_Use_Config_INI(self, use):
        """ Congure F Initialization to use (default) or skip config.ini file. When skipping config.ini it is
            necessary to use F_CongFileLoad or F_Set_Cong_Value_By_Name functions to congure the adapter."""
        return self.__sel_dll.F_Use_Config_INI(use)

    def F_Get_Config_Name_List(self, index):
        """ Iterate over conguration parameter names, starting from index 0. Increase index until return is
            null character. Will return a pointer to a character array with name of parameter or null character
            once index is too high.
            Requires the target FPA to be opened using F OpenInstancesAndFPAs and initialized using F_Initialization."""
        return self.__sel_dll.F_Get_Config_Name_List(index)

    def F_Get_Config_Value_By_Name(self, name, typ):
        """ Access the current FPA's conguration. Use F_Get_Cong_Name_List to get conguration parameter 
            names, then use them as input to this function. Select input type to validate name, and get
            current, minimum, maximum, and default values.
            Because normal return values from this function can be any value, from 0 to 0xFFFFFFFF,
            it is important to pick a correct fpa index (use F Check FPA index to validate) and validate the
            conguration parameter name rst (using CONFSEL VALIDATE), before retrieving the current,
            minimum, maximum or default values. Unless all FPAs are to be congured identically, avoid
            using FPA index 0 to prevent confusion in results (when results don't match using FPA index 0
            (all FPAs) the result is simply FPA UNMATCHED RESULTS (-1)).
            Requires the target FPA to be opened using F OpenInstancesAndFPAs and initialized using F_Initialization. """
        return self.__sel_dll.F_Get_Config_Value_By_Name(name, typ)

    def F_Set_Config_Value_By_Name(self, name, new_value):
        """ Set the current FPA's conguration. Use F_Get_Cong_Name_List to get conguration parameter
            names, then use them as input to this function. New value for conguration parameter
            will be trimmed by minimum and maximum values allowed for this parameter. If not initialized, 
            this parameter will have a default value. Double check parameters after conguration using F_Get_Cong_Value_By_Name.
            Requires the target FPA to be opened using F OpenInstancesAndFPAs and initialized using F_Initialization."""
        return self.__sel_dll.F_Set_Config_Value_By_Name(name, new_value)

    def F_Get_MCU_Name_list(self, typ, index):
        """ Obtain a list of supported vendors, families, groups, and MCUs for the specied FPA. Use this
            function in conjunction with F_Set_MCU_Family_Group to get all supported MCU names. First
            obtain the names of supported vendors, then select one vendor. Then, for the selected vendor
            obtain the names of supported families, then select one family, etc.
            The MCU names obtained from this function are the exact names used to congure the FPA
            using the function F_Set_MCU_Name for the target device you want to program."""
        func = self.__sel_dll.F_Get_MCU_Name_list
        func.restype = c_char_p
        func.argtypes = [c_int, c_int] 
        #return self.__sel_dll.F_Get_MCU_Name_list(typ, index)
        return func(typ, index)

    def F_Set_MCU_Family_Group(self, typ, index):
        """ Select vendor, family, group and MCU index. Use this function in conjunction with
            F_Get_MCU_Name_list to get all supported MCU names. First obtain the names of supported
            vendors, then select one vendor. Then, for the selected vendor obtain the names of supported
            families, then select one family, etc.
            Use MCU names to congure the FPA using the function F Set MCU Name. Use this function
            only to obtain the list of supported MCU names, not to congure the FPA. When NEW MCUs are
            supported then these indices will change."""
        return self.__sel_dll.F_Set_MCU_Family_Group(typ, index)

    def F_DLLTypeVer(self):
        """ Get information about DLL software type and version. This function returns integer number with
            DLL ID and software revision version and copies a text message to the report message buffer about
            DLL ID and software revision. Text content can read using F_ReportMessage, F_Report_Message,
            or F_GetReportMessageChar functions."""
        return self.__sel_dll.F_DLLTypeVer()
    
    def F_ConfigFileLoad(self, filename):
        """ Modify programmer's conguration according to data taken from specied configuration file.
            This setup will override previous values and leave omitted parameters untouched. When loaded
            for the first time, unspecified conguration parameters take on default values as given by
            F_Get_Cong_Value_By_Name. When calling this function multiple times with different configuration
            files, make sure to specify all relevant configuration parameters.
            A configuration le can be created using the FlashPro GUI software, or use the functions
            F_Get_Cong_Name_List, F_Get_Config_Value By Name to extract a list of configuration names
            and values. The F_Set_Cong_Value By Name function can be used to set conguration parameters
            instead of F_ConfigFileLoad.
            Requires the target FPA to be opened using F OpenInstancesAndFPAs and initialized using F_Initialization."""
        return self.__sel_dll.F_ConfigFileLoad(filename)
        
    def F_Power_Target(self, on_off):
        """ Enable power from FPA to target device. For this function to have its desired effect, configuration
            settings PowerFromFpaEn must be set to 1. If PowerFromFpaEn is set to 0, then power from FPA
            is always off regardless of this function's input parameter."""
        return self.__sel_dll.F_Power_Target(on_off)

    def F_Get_Targets_Vcc(self):
        """ Select FPA to perform operation on using F_Set_FPA_index, index 1 to 64.Use index 0 to
            perform operation on all FPAs (if results differ, use F_LastStatus to get individual results)."""
        return self.__sel_dll.F_Get_Targets_Vcc()

    def F_Set_fpa_io_state(self, jtag, reset, vcc_on):
        """ Set state of the Reset, Vcc and JTAG lines after programming is finished."""
        return self.__sel_dll.F_Set_fpa_io_state(jtag, reset, vcc_on)
    

    # DATA BUFFER FUNCTIONS

    def F_ReadCodeFile(self, filename):
        """ Read code from le and store in internal FPA buffer to be used in programming. Only code
            data that ts within the target MCUs memory space will be read, and the rest will be discarded.
            Therefore, it is necessary to congure the FPA for the correct MCU rst, then load the code le.
            Unspecied locations within the code  file are set to the target MCU's default empty value (for most
            MCUs it is 0xFF, for some it is 0x00)."""
        return self.__sel_dll.F_ReadCodeFile(filename)

    def F_AppendCodeFile(self, filename):
        """ Append code from selected le to internal FPA buffer to be used in programming. Only code data
            that ts within the target MCUs memory space will be read, and the rest will be discarded.
            Already read code data is not overwritten by newly appended le. Existing code data locations
            that contain the default empty value (for most MCUs it is 0xFF, for some it is 0x00) can be
            overwritten if the OverwriteEmptyValues option is enabled in the Preferences le, prefer.ini (read
            from current working directory). By default, no overwrites are performed."""
        return self.__sel_dll.F_AppendCodeFile(filename)
    
    def F_Get_CodeCS(self, dest):
        """ Read code from selected buffer and calculate check sum."""
        return self.__sel_dll.F_Get_CodeCS(dest)

    def F_Clr_Code_Buffer(self):
        """ Clear contents of internal code buffer. """
        return self.__sel_dll.F_Clr_Code_Buffer()

    def F_Put_Byte_to_Code_Buffer(self, addr, data):
        """ Write to internal code buffer. Can be used instead of or in conjunction with the F_ReadCodeFile
            function. When starting from scratch, use the F_Clr_Code_Buffer function to clear the internal code buffer."""
        return self.__sel_dll.F_Put_Byte_to_Code_Buffer(addr, data)
    
    def F_Get_Byte_from_Code_Buffer(self, addr):
        """ Read from internal code buffer. Can be used in conjunction with F_Put_Byte_to_Code_Buffer to
            verify writes to internal code buffer."""
        return self.__sel_dll.F_Get_Byte_from_Code_Buffer(addr)

    def F_Put_Byte_to_Buffer(self, addr, data):
        """ Write byte to temporary Write Data Buffer."""
        return self.__sel_dll.F_Put_Byte_to_Buffer(addr, data)

    def F_Get_Byte_from_Buffer(self, addr):
        """ Read from temporary Read Data Buffer."""
        return self.__sel_dll.F_Get_Byte_from_Buffer(addr)


    # ENCAPSULATED FUNCTIONS

    def F_AutoProgram(self, mode):
        """ Perform full programming sequence on selected FPA. Power on (optional), open communication,
            erase, blank check, program, verify and lock device (optional). Exact behavior of each action is
            controlled through conguration settings (i.e. PowerFromFpaEn has to be enabled to power target
            device from FPA, otherwise outside power source is required)."""
        return self.__sel_dll.F_AutoProgram(mode)

    def F_Verify_Access_to_MCU(self):
        """ Determine if FPA has successfully established communication with the target device. This step
            requires a working connection, a read of the processor ID, memory size, and 
            ash size checked against internal records (except when target MCU doesn't have this meta-data available), and
            calibration of internal CPU clock. If these steps match expected values the operation will succeed.
            This function is internally called by all other encapsulated functions at startup before they
            proceed to perform any other operations (except one). Therefore if this function fails, other encapsulated 
            functions will invariably fail as well. The exception is F Clear Locked Device which uses a
            custom procedure to unlock the MCU (if supported by vendor)."""
        return self.__sel_dll.F_Verify_Access_to_MCU()

    def F_Memory_Erase(self, mode):
        """ Erase memory specied in conguration options. Depending on settings, this operation will erase
            all or part of target device's memory contents. Use Retain Data settings to preserve calibration
            or custom data. Retain Data settings only apply to encapsulated functions (F_AutoProgram, this
            function, etc.) custom writes using Data Buffer or Sequential instructions will not retain data."""
        return self.__sel_dll.F_Memory_Erase(mode)
    
    def F_Memory_Blank_Check(self):
        """ Blank check memory specied in conguration options. Depending on settings, this operation will
            blank check all or part of target device's memory contents."""
        return self.__sel_dll.F_Memory_Blank_Check()

    def F_Memory_Write(self, mode):
        """ Write contents of Code Buffer to target device's memory. The Code Buffer is usually written
            from a code le using F ReadCodeFile, but can be modied with custom instructions such as
            F_Put_Byte_to_Code_Buffer. An enabled conguration option, CodeFileReload, will cause this
            function to always reload code le before proceeding. This setting can be useful if code le is
            frequently modied, but it can cause issues if custom instructions are used to modify the code
            buffer before programming, as their modications will be discarded."""
        return self.__sel_dll.F_Memory_Write(mode)

    def F_Memory_Verify(self, mode):
        """ Verify target device's memory contents against Code Buffer. The sections compared depend
            on conguration settings, and are identical to sections of memory being programmed using
            F_Memory_Write. For common behavior, do not modify settings between calling F_Memory_Write
            and this function."""
        return self.__sel_dll.F_Memory_Verify(mode)

    def F_Memory_Read(self):
        """ Read target device's memory contents and store in Read Data Buffer.
            The sections of memory read depend on conguration settings, FlashReadModeIndex. Use
            F_Get_Byte_from_Buffer to access the contents of this buffer after this function has completed."""
        return self.__sel_dll.F_Memory_Read()

    def F_Lock_MCU(self):
        """ This function can write memory protection bits and disable debug access. Performs the functions
            F_Write_Locking_Registers and F_Write_Debug_Register together. If the former fails, the latter will
            not execute (locking bits are veried rst before debug access is disabled).
            This procedure is not supported by all target devices. Consult individual device manuals or the
            FlashPro GUI, Setup, Memory Protection window which will indicate which bits can be set."""
        return self.__sel_dll.F_Lock_MCU()

    def F_Clear_Locked_Device(self):
        """ Will attempt to return all memory to factory settings. Usually requires a mass erase of the memory
            contents. This procedure can be used to unlock memory protection or enable debug access after
            the F_Lock_MCU procedure.
            This procedure is not supported by all target devices. Consult individual device manuals or the
            FlashPro GUI, Setup, Memory Protection window which will indicate which bits can be reset."""
        return self.__sel_dll.F_Clear_Locked_Device()

    # SEQUENTIAL FUNCTIONS

    def F_Open_Target_Device(self):
        """ The FPA will establish communication with the target device. This step requires a working connection, 
            a read of the processor ID, memory size, and ash size checked against internal records
            (except when target MCU doesn't have this meta-data available), and calibration of internal CPU
            clock. If these steps match expected values the operation will succeed.
            This function is identical to F_Verify_Access to MCU except that the target device is not closed
            when this function concludes (F_Open_Target_Device keeps the target device open to accept further
            instructions listed below)."""
        return self.__sel_dll.F_Open_Target_Device()

    def F_Close_Target_Device(self):
        """ Terminates communication between FPA and target device. Sets reset line to GND and communication lines (TMS, TCK, TDI) 
            to conguration setting ReleaseJtagState."""
        return self.__sel_dll.F_Close_Target_Device()

    def F_Segment_Erase(self, address):
        """ Read specied number of bytes from MCU flash or OTP to temporary Read Data Buffer"""
        return self.__sel_dll.F_Segment_Erase(address)

    def F_Sectors_Blank_Check(self, start_address, stop_address):
        """ Blank check part or all of the target device's memory. """
        return self.__sel_dll.F_Sectors_Blank_Check(start_address, stop_address)

    def F_Get_Sector_Size(self, address):
        """ Will return the segment/sector size for memory region that contains given address. This function
            will access the FPA's meta-data for the target device specied in conguration options.
            No communication with target device actually takes place."""
        return self.__sel_dll.F_Get_Sector_Size(address)
    
    def F_Copy_Buffer_to_Flash(self, start_address, size):
        """ Blank check part or all of the target device's memory."""
        return self.__sel_dll.F_Copy_Buffer_to_Flash(start_address, size)

    def F_Copy_Flash_to_Buffer(self, start_address, size):
        """ Read specied number of bytes from MCU flash or OTP to temporary Read Data Buffer"""
        return self.__sel_dll.F_Copy_Flash_to_Buffer(start_address, size)
    
    def F_Write_Byte_to_RAM(self, address, data):
        """ Use the FPA to write a byte to the target device's RAM. This write is not veried (no read performed). Use F Read Byte to verify."""
        return self.__sel_dll.F_Write_Byte_to_RAM(address, data)

    def F_Write_Word16_to_RAM(self, address, data):
        """ Use the FPA to write two bytes to the target device's RAM. This write is not veried (no read
            performed). Use F Read Word16 to verify."""
        return self.__sel_dll.F_Write_Word16_to_RAM(address, data)

    def F_Write_Word32_to_RAM(self, address, data):
        """ Use the FPA to write four bytes to the target device's RAM. This write is not veried (no read
            performed). Use F Read Word32 to verify."""
        return self.__sel_dll.F_Write_Word32_to_RAM(address, data)
    
    def F_Write_Bytes_Block_to_RAM(self, address, size, data):
        """ Use the FPA to write a block of bytes to RAM. Currently this function only writes a block of bytes
            with a size that is a multiple of four, to a 32-bit word aligned address. These writes are not veried
            (no reads performed). Use F Read Bytes Block to verify."""
        return self.__sel_dll.F_Write_Bytes_Block_to_RAM(address, size, data)
    
    def F_Read_Byte(self, address):
        """ Use the FPA to read one byte from any address (RAM, flash, etc.)."""
        return self.__sel_dll.F_Read_Byte(address)

    def F_Read_Word16(self, address):
        """ Use the FPA to read two bytes from any address (RAM, flash, etc.)."""
        return self.__sel_dll.F_Read_Word16(address)

    def F_Read_Word32(self, address):
        """Use the FPA to read four bytes from any address (RAM, flash, etc.)."""
        return self.__sel_dll.F_Read_Word32(address)

    def F_Read_Bytes_Block(self, address, size, data):
        """ Use the FPA to read a block of bytes from any address (RAM, flash, etc.)."""
        return self.__sel_dll.F_Read_Bytes_Block(address, size, data)

    def F_Set_PC_and_RUN(self, pc_addr):
        """ Start target MCU from specied program counter (PC). There should be a valid instruction at the target address."""
        return self.__sel_dll.F_Set_PC_and_RUN(pc_addr)

    def F_Write_Locking_Registers(self):
        """ Write to non-volatile memory protection bits and/or persistent user-data registers (if supported).
            Does not block communication over debug interface, but can disable writing and/or reading from
            target device's memory. These bits can be reset to factory settings using the F_Clear_Locked_Device
            function if the device supports it. To write protection bits, enable the WriteLockingBitsEn configuration option."""
        return self.__sel_dll.F_Write_Locking_Registers()
    
    def F_Write_Debug_Register(self):
        """ Write to non-volatile protection bits to disable communication over debug interface. These bits can
            be reset to factory settings using the F Clear Locked Device function if the device supports it. To
            disable debug access, enable WriteLockingBitsEn and (vendor)_USER_DBG_WrEn conguration options."""
        return self.__sel_dll.F_Write_Debug_Register()
    
    def F_Get_MCU_Data(self, typ):
        """ Return the MCU ID, flash size, or RAM size read by the FPA to identify a specic target device if supported by vendor. """
        return self.__sel_dll.F_Get_MCU_Data(typ)
